# Copyright (c) 2015 OpenStack Foundation.
# All Rights Reserved.

#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

REGION_ONE_PUBLIC_KEYSTONE_ENDPOINT_ID = '8cd4b957090f4ca5842a22e9a74099cd'


PROJECT_SCOPED_TOKEN_FIXTURE = {
    'token': {
        'methods': ['password'],
        'roles': [
            {'id': 'f03fda8f8a3249b2a70fb1f176a7b631', 'name': 'role1'},
            {'id': 'f03fda8f8a3249b2a70fb1f176a7b631', 'name': 'role2'},
        ],
        'issued_at': '2002-01-18T21:14:07Z',
        'expires_at': '2038-01-18T21:14:07Z',
        'project': {
            'id': 'tenant_id1',
            'domain': {'id': 'domain_id1', 'name': 'domain_name1'},
            'enabled': True,
            'description': 'no description available',
            'name': 'tenant_name1',
        },
        'catalog': [
            {
                'endpoints': [
                    {
                        'id': '3b5e554bcf114f2483e8a1be7a0506d1',
                        'interface': 'admin',
                        'url': 'http://127.0.0.1:8776/v1/'
                        + '64b6f3fbcc53435e8a60fcf89bb6617a',
                        'region': 'regionOne',
                    },
                    {
                        'id': '54abd2dc463c4ba4a72915498f8ecad1',
                        'interface': 'internal',
                        'url': 'http://127.0.0.1:8776/v1/'
                        + '64b6f3fbcc53435e8a60fcf89bb6617a',
                        'region': 'regionOne',
                    },
                    {
                        'id': '70a7efa4b1b941968357cc43ae1419ee',
                        'interface': 'public',
                        'url': 'http://127.0.0.1:8776/v1/'
                        + '64b6f3fbcc53435e8a60fcf89bb6617a',
                        'region': 'regionOne',
                    },
                ],
                'id': '5707c3fc0a294703a3c638e9cf6a6c3a',
                'type': 'volume',
                'name': 'volume',
            },
            {
                'endpoints': [
                    {
                        'id': '92217a3b95394492859bc49fd474382f',
                        'interface': 'admin',
                        'url': 'http://127.0.0.1:9292/v1',
                        'region': 'regionOne',
                    },
                    {
                        'id': 'f20563bdf66f4efa8a1f11d99b672be1',
                        'interface': 'internal',
                        'url': 'http://127.0.0.1:9292/v1',
                        'region': 'regionOne',
                    },
                    {
                        'id': '375f9ba459a447738fb60fe5fc26e9aa',
                        'interface': 'public',
                        'url': 'http://127.0.0.1:9292/v1',
                        'region': 'regionOne',
                    },
                ],
                'id': '15c21aae6b274a8da52e0a068e908aac',
                'type': 'image',
                'name': 'glance',
            },
            {
                'endpoints': [
                    {
                        'id': 'edbd9f50f66746ae9ed11dc3b1ae35da',
                        'interface': 'admin',
                        'url': 'http://127.0.0.1:8774/v1.1/'
                        + '64b6f3fbcc53435e8a60fcf89bb6617a',
                        'region': 'regionOne',
                    },
                    {
                        'id': '9e03c46c80a34a159cb39f5cb0498b92',
                        'interface': 'internal',
                        'url': 'http://127.0.0.1:8774/v1.1/'
                        + '64b6f3fbcc53435e8a60fcf89bb6617a',
                        'region': 'regionOne',
                    },
                    {
                        'id': '1df0b44d92634d59bd0e0d60cf7ce432',
                        'interface': 'public',
                        'url': 'http://127.0.0.1:8774/v1.1/'
                        + '64b6f3fbcc53435e8a60fcf89bb6617a',
                        'region': 'regionOne',
                    },
                ],
                'id': '2f404fdb89154c589efbc10726b029ec',
                'type': 'compute',
                'name': 'nova',
            },
            {
                'endpoints': [
                    {
                        'id': 'a4501e141a4b4e14bf282e7bffd81dc5',
                        'interface': 'admin',
                        'url': 'http://127.0.0.1:35357/v3',
                        'region': 'RegionOne',
                    },
                    {
                        'id': '3d17e3227bfc4483b58de5eaa584e360',
                        'interface': 'internal',
                        'url': 'http://127.0.0.1:35357/v3',
                        'region': 'RegionOne',
                    },
                    {
                        'id': REGION_ONE_PUBLIC_KEYSTONE_ENDPOINT_ID,
                        'interface': 'public',
                        'url': 'http://127.0.0.1:5000/v3',
                        'region': 'RegionOne',
                    },
                ],
                'id': 'c5d926d566424e4fba4f80c37916cde5',
                'type': 'identity',
                'name': 'keystone',
            },
        ],
        'user': {
            'domain': {'id': 'domain_id1', 'name': 'domain_name1'},
            'name': 'user_name1',
            'id': 'user_id1',
        },
    }
}

SYSTEM_SCOPED_TOKEN_FIXTURE = {
    'token': {
        'methods': ['password'],
        'expires_at': '2038-01-18T21:14:07Z',
        'issued_at': '2000-01-18T21:14:07Z',
        'roles': [
            {'id': '41b1af9bb39241e8b8b79fae5906abcc', 'name': 'role1'},
            {'id': 'ac9add6b3c5a46dcaaf21390c4657949', 'name': 'role2'},
        ],
        'system': {'all': True},
        'catalog': [
            {
                'endpoints': [
                    {
                        'id': '3b5e554bcf114f2483e8a1be7a0506d1',
                        'interface': 'admin',
                        'url': 'http://127.0.0.1:8776/v1/'
                        + '64b6f3fbcc53435e8a60fcf89bb6617a',
                        'region': 'regionOne',
                    },
                    {
                        'id': '54abd2dc463c4ba4a72915498f8ecad1',
                        'interface': 'internal',
                        'url': 'http://127.0.0.1:8776/v1/'
                        + '64b6f3fbcc53435e8a60fcf89bb6617a',
                        'region': 'regionOne',
                    },
                    {
                        'id': '70a7efa4b1b941968357cc43ae1419ee',
                        'interface': 'public',
                        'url': 'http://127.0.0.1:8776/v1/'
                        + '64b6f3fbcc53435e8a60fcf89bb6617a',
                        'region': 'regionOne',
                    },
                ],
                'id': '5707c3fc0a294703a3c638e9cf6a6c3a',
                'type': 'volume',
                'name': 'volume',
            },
            {
                'endpoints': [
                    {
                        'id': '92217a3b95394492859bc49fd474382f',
                        'interface': 'admin',
                        'url': 'http://127.0.0.1:9292/v1',
                        'region': 'regionOne',
                    },
                    {
                        'id': 'f20563bdf66f4efa8a1f11d99b672be1',
                        'interface': 'internal',
                        'url': 'http://127.0.0.1:9292/v1',
                        'region': 'regionOne',
                    },
                    {
                        'id': '375f9ba459a447738fb60fe5fc26e9aa',
                        'interface': 'public',
                        'url': 'http://127.0.0.1:9292/v1',
                        'region': 'regionOne',
                    },
                ],
                'id': '15c21aae6b274a8da52e0a068e908aac',
                'type': 'image',
                'name': 'glance',
            },
            {
                'endpoints': [
                    {
                        'id': 'edbd9f50f66746ae9ed11dc3b1ae35da',
                        'interface': 'admin',
                        'url': 'http://127.0.0.1:8774/v1.1/'
                        + '64b6f3fbcc53435e8a60fcf89bb6617a',
                        'region': 'regionOne',
                    },
                    {
                        'id': '9e03c46c80a34a159cb39f5cb0498b92',
                        'interface': 'internal',
                        'url': 'http://127.0.0.1:8774/v1.1/'
                        + '64b6f3fbcc53435e8a60fcf89bb6617a',
                        'region': 'regionOne',
                    },
                    {
                        'id': '1df0b44d92634d59bd0e0d60cf7ce432',
                        'interface': 'public',
                        'url': 'http://127.0.0.1:8774/v1.1/'
                        + '64b6f3fbcc53435e8a60fcf89bb6617a',
                        'region': 'regionOne',
                    },
                ],
                'id': '2f404fdb89154c589efbc10726b029ec',
                'type': 'compute',
                'name': 'nova',
            },
            {
                'endpoints': [
                    {
                        'id': 'a4501e141a4b4e14bf282e7bffd81dc5',
                        'interface': 'admin',
                        'url': 'http://127.0.0.1:35357/v3',
                        'region': 'RegionOne',
                    },
                    {
                        'id': '3d17e3227bfc4483b58de5eaa584e360',
                        'interface': 'internal',
                        'url': 'http://127.0.0.1:35357/v3',
                        'region': 'RegionOne',
                    },
                    {
                        'id': REGION_ONE_PUBLIC_KEYSTONE_ENDPOINT_ID,
                        'interface': 'public',
                        'url': 'http://127.0.0.1:5000/v3',
                        'region': 'RegionOne',
                    },
                ],
                'id': 'c5d926d566424e4fba4f80c37916cde5',
                'type': 'identity',
                'name': 'keystone',
            },
        ],
        'user': {
            'domain': {'id': 'domain_id1', 'name': 'domain_name1'},
            'name': 'user_name1',
            'id': 'user_id1',
        },
    }
}



#!/usr/bin/env python3

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import MutableMapping
from typing import Any

from oslo_serialization import jsonutils

from oslo_config import cfg
from oslo_policy import _checks
from oslo_policy import opts
from oslo_policy import policy


class FakeEnforcer:
    def __init__(self, rules: policy.Rules, config: str | None) -> None:
        self.rules = rules
        self.conf = None

        if config:
            self.conf = cfg.ConfigOpts()

            for group, options in opts.list_opts():
                self.conf.register_opts(options, group)

            self.conf([f'--config-file={config}'])


def _try_rule(
    key: str,
    rule: _checks.BaseCheck,
    target: _checks.TargetT,
    access_data: MutableMapping[str, Any],
    o: FakeEnforcer,
) -> None:
    try:
        result = rule(
            target,
            access_data,
            o,  # type: ignore
            current_rule=key,
        )
        if result:
            print(f'passed: {key}')
        else:
            print(f'failed: {key}')
    except Exception as e:
        print(e)
        print(f'exception: {rule}')


def flatten(
    d: MutableMapping[str, Any], parent_key: str = ''
) -> dict[str, Any]:
    """Flatten a nested dictionary

    Converts a dictionary with nested values to a single level flat
    dictionary, with dotted notation for each key.

    """
    items: list[tuple[str, Any]] = []
    for k, v in d.items():
        new_key = parent_key + '.' + k if parent_key else k
        if isinstance(v, MutableMapping):
            items.extend(flatten(v, new_key).items())
        else:
            items.append((new_key, v))
    return dict(items)


def tool(
    policy_file: str,
    access_file: str,
    apply_rule: str | None,
    is_admin: bool = False,
    target_file: str | None = None,
    enforcer_config: str | None = None,
) -> None:
    with open(access_file, 'rb', 0) as a:
        access = a.read()

    access_data = jsonutils.loads(access)['token']
    access_data['roles'] = [role['name'] for role in access_data['roles']]
    access_data['user_id'] = access_data['user']['id']
    if access_data.get('project'):
        access_data['project_id'] = access_data['project']['id']
    if access_data.get('system'):
        access_data['system_scope'] = 'all'
    access_data['is_admin'] = is_admin

    with open(policy_file, 'rb', 0) as p:
        policy_data = p.read()

    rules = policy.Rules.load(policy_data, 'default')

    enforcer = FakeEnforcer(rules, enforcer_config)

    if target_file:
        with open(target_file, 'rb', 0) as t:
            target = t.read()

        target_data = flatten(jsonutils.loads(target))
    else:
        target_data = {'user_id': access_data['user']['id']}
        if access_data.get('project_id'):
            target_data['project_id'] = access_data['project_id']

    if apply_rule:
        key = apply_rule
        rule = rules[apply_rule]
        _try_rule(key, rule, target_data, access_data, enforcer)
        return

    for key, rule in sorted(rules.items()):
        if ':' in key:
            _try_rule(key, rule, target_data, access_data, enforcer)


def main() -> None:
    conf = cfg.ConfigOpts()

    conf.register_cli_opt(
        cfg.StrOpt('policy', required=True, help='path to a policy file.')
    )

    conf.register_cli_opt(
        cfg.StrOpt(
            'access',
            required=True,
            help=(
                'path to a file containing OpenStack Identity API '
                'access info in JSON format.'
            ),
        )
    )

    conf.register_cli_opt(
        cfg.StrOpt(
            'target',
            help=(
                'path to a file containing custom target info in '
                'JSON format. This will be used to evaluate the policy with.'
            ),
        )
    )

    conf.register_cli_opt(cfg.StrOpt('rule', help='rule to test.'))

    conf.register_cli_opt(
        cfg.BoolOpt(
            'is_admin',
            help=(
                'set is_admin=True on the credentials used for the evaluation.'
            ),
            default=False,
        )
    )

    conf.register_cli_opt(
        cfg.StrOpt(
            'enforcer_config',
            help='configuration file for the oslopolicy-checker enforcer',
        )
    )

    conf()

    tool(
        conf.policy,
        conf.access,
        conf.rule,
        conf.is_admin,
        conf.target,
        conf.enforcer_config,
    )


if __name__ == '__main__':
    main()



# Copyright (c) 2015 OpenStack Foundation.
# All Rights Reserved.

#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from typing import Any
from unittest import mock

from oslotest import base as test_base

from oslo_policy import _checks
from oslo_policy import policy
from oslo_policy.tests import base
from oslo_policy.tests import token_fixture


class CheckRegisterTestCase(test_base.BaseTestCase):
    @mock.patch.object(_checks, 'registered_checks', {})
    def test_register_check(self):
        class TestCheck(_checks.Check):
            pass

        _checks.register('spam', TestCheck)

        self.assertEqual({'spam': TestCheck}, _checks.registered_checks)

    @mock.patch.object(_checks, 'registered_checks', {})
    def test_register_check_decorator(self):
        @_checks.register('spam')
        class TestCheck(_checks.Check):
            pass

        self.assertEqual({'spam': TestCheck}, _checks.registered_checks)


class RuleCheckTestCase(base.PolicyBaseTestCase):
    def test_rule_missing(self):
        self.enforcer.rules = policy.Rules({})
        check = _checks.RuleCheck('rule', 'spam')

        self.assertFalse(check({}, {}, self.enforcer))

    def test_rule_false(self):
        self.enforcer.rules = policy.Rules({'spam': _BoolCheck(False)})
        check = _checks.RuleCheck('rule', 'spam')

        self.assertFalse(check({}, {}, self.enforcer))

    def test_rule_true(self):
        self.enforcer.rules = policy.Rules({'spam': _BoolCheck(True)})
        check = _checks.RuleCheck('rule', 'spam')

        self.assertTrue(check({}, {}, self.enforcer))


class RoleCheckTestCase(base.PolicyBaseTestCase):
    def test_accept(self):
        check = _checks.RoleCheck('role', 'sPaM')

        self.assertTrue(check({}, {'roles': ['SpAm']}, self.enforcer))

    def test_reject(self):
        check = _checks.RoleCheck('role', 'spam')

        self.assertFalse(check({}, {'roles': []}, self.enforcer))

    def test_format_value(self):
        check = _checks.RoleCheck('role', '%(target.role.name)s')

        target_dict: dict[str, Any] = {'target.role.name': 'a'}
        cred_dict = {'user': 'user', 'roles': ['a', 'b', 'c']}
        self.assertTrue(check(target_dict, cred_dict, self.enforcer))

        target_dict = {'target.role.name': 'd'}
        self.assertFalse(check(target_dict, cred_dict, self.enforcer))

        target_dict = {'target': {'role': {}}}
        self.assertFalse(check(target_dict, cred_dict, self.enforcer))

    def test_no_roles_case(self):
        check = _checks.RoleCheck('role', 'spam')

        self.assertFalse(check({}, {}, self.enforcer))


class GenericCheckTestCase(base.PolicyBaseTestCase):
    def test_no_cred(self):
        check = _checks.GenericCheck('name', '%(name)s')

        self.assertFalse(check({'name': 'spam'}, {}, self.enforcer))

    def test_cred_mismatch(self):
        check = _checks.GenericCheck('name', '%(name)s')

        self.assertFalse(
            check({'name': 'spam'}, {'name': 'ham'}, self.enforcer)
        )

    def test_accept(self):
        check = _checks.GenericCheck('name', '%(name)s')

        self.assertTrue(
            check({'name': 'spam'}, {'name': 'spam'}, self.enforcer)
        )

    def test_no_key_match_in_target(self):
        check = _checks.GenericCheck('name', '%(name)s')

        self.assertFalse(
            check({'name1': 'spam'}, {'name': 'spam'}, self.enforcer)
        )

    def test_constant_string_mismatch(self):
        check = _checks.GenericCheck("'spam'", '%(name)s')

        self.assertFalse(check({'name': 'ham'}, {}, self.enforcer))

    def test_constant_string_accept(self):
        check = _checks.GenericCheck("'spam'", '%(name)s')

        self.assertTrue(check({'name': 'spam'}, {}, self.enforcer))

    def test_constant_literal_mismatch(self):
        check = _checks.GenericCheck('True', '%(enabled)s')

        self.assertFalse(check(dict(enabled=False), {}, self.enforcer))

    def test_constant_literal_accept(self):
        check = _checks.GenericCheck('True', '%(enabled)s')

        self.assertTrue(check(dict(enabled=True), {}, self.enforcer))

    def test_deep_credentials_dictionary_lookup(self):
        check = _checks.GenericCheck('a.b.c.d', 'APPLES')

        credentials = {'a': {'b': {'c': {'d': 'APPLES'}}}}

        self.assertTrue(check({}, credentials, self.enforcer))

    def test_missing_credentials_dictionary_lookup(self):
        credentials = {'a': 'APPLES', 'o': {'t': 'ORANGES'}}

        # First a valid check - rest of case is expecting failures
        # Should prove the basic credentials structure before we test
        # for failure cases.
        check = _checks.GenericCheck('o.t', 'ORANGES')
        self.assertTrue(check({}, credentials, self.enforcer))

        # Case where final key is missing
        check = _checks.GenericCheck('o.v', 'ORANGES')
        self.assertFalse(check({}, credentials, self.enforcer))

        # Attempt to access key under a missing dictionary
        check = _checks.GenericCheck('q.v', 'APPLES')
        self.assertFalse(check({}, credentials, self.enforcer))

    def test_single_entry_in_list_accepted(self):
        check = _checks.GenericCheck('a.b.c.d', 'APPLES')
        credentials = {'a': {'b': {'c': {'d': ['APPLES']}}}}
        self.assertTrue(check({}, credentials, self.enforcer))

    def test_multiple_entry_in_list_accepted(self):
        check = _checks.GenericCheck('a.b.c.d', 'APPLES')
        credentials = {
            'a': {'b': {'c': {'d': ['Bananas', 'APPLES', 'Grapes']}}}
        }
        self.assertTrue(check({}, credentials, self.enforcer))

    def test_multiple_entry_in_nested_list_accepted(self):
        check = _checks.GenericCheck('a.b.c.d', 'APPLES')
        credentials = {
            'a': {'b': [{'c': {'d': ['BANANAS', 'APPLES', 'GRAPES']}}, {}]}
        }
        self.assertTrue(check({}, credentials, self.enforcer))

    def test_multiple_entries_one_matches(self):
        check = _checks.GenericCheck(
            'token.catalog.endpoints.id',
            token_fixture.REGION_ONE_PUBLIC_KEYSTONE_ENDPOINT_ID,
        )
        credentials = token_fixture.PROJECT_SCOPED_TOKEN_FIXTURE
        self.assertTrue(check({}, credentials, self.enforcer))

    def test_generic_role_check_matches(self):
        check = _checks.GenericCheck('token.roles.name', 'role1')
        credentials = token_fixture.PROJECT_SCOPED_TOKEN_FIXTURE
        self.assertTrue(check({}, credentials, self.enforcer))

    def test_generic_missing_role_does_not_matches(self):
        check = _checks.GenericCheck('token.roles.name', 'missing')
        credentials = token_fixture.PROJECT_SCOPED_TOKEN_FIXTURE
        self.assertFalse(check({}, credentials, self.enforcer))

    def test_multiple_nested_lists_accepted(self):
        check = _checks.GenericCheck('a.b.c.d', 'APPLES')
        credentials = {
            'a': {
                'b': [
                    {'a': ''},
                    {'c': {'d': ['BANANAS', 'APPLES', 'GRAPES']}},
                    {},
                ]
            }
        }
        self.assertTrue(check({}, credentials, self.enforcer))

    def test_entry_not_in_list_rejected(self):
        check = _checks.GenericCheck('a.b.c.d', 'APPLES')
        credentials = {'a': {'b': {'c': {'d': ['PEACHES', 'PEARS']}}}}
        self.assertFalse(check({}, credentials, self.enforcer))


class FalseCheckTestCase(base.PolicyBaseTestCase):
    def test_str(self):
        check = _checks.FalseCheck()

        self.assertEqual('!', str(check))

    def test_call(self):
        check = _checks.FalseCheck()

        self.assertFalse(check({}, {}, self.enforcer))


class TrueCheckTestCase(base.PolicyBaseTestCase):
    def test_str(self):
        check = _checks.TrueCheck()

        self.assertEqual('@', str(check))

    def test_call(self):
        check = _checks.TrueCheck()

        self.assertTrue(check({}, {}, self.enforcer))


class CheckForTest(_checks.Check):
    def __call__(self, target, creds, enforcer, current_rule=None):
        pass


class CheckTestCase(base.PolicyBaseTestCase):
    def test_init(self):
        check = CheckForTest('kind', 'match')

        self.assertEqual('kind', check.kind)
        self.assertEqual('match', check.match)

    def test_str(self):
        check = CheckForTest('kind', 'match')

        self.assertEqual('kind:match', str(check))


class NotCheckTestCase(base.PolicyBaseTestCase):
    def test_init(self):
        foo = CheckForTest('kind', 'match')
        check = _checks.NotCheck(foo)

        self.assertEqual(foo, check.rule)

    def test_str(self):
        rule = CheckForTest('kind', 'match')
        check = _checks.NotCheck(rule)

        self.assertEqual('not kind:match', str(check))

    def test_call_true(self):
        rule = _checks.TrueCheck()
        check = _checks.NotCheck(rule)
        self.assertFalse(check({}, {}, self.enforcer))

    def test_call_false(self):
        rule = _checks.FalseCheck()
        check = _checks.NotCheck(rule)
        self.assertTrue(check({}, {}, self.enforcer))

    def test_rule_takes_current_rule(self):
        results = []

        class TestCheck(_checks.Check):
            def __call__(self, target, cred, enforcer, current_rule=None):
                results.append((target, cred, enforcer, current_rule))
                return True

        check = _checks.NotCheck(TestCheck('kind', 'match'))

        self.assertFalse(check({}, {}, self.enforcer, current_rule='a_rule'))
        self.assertEqual(
            [({}, {}, self.enforcer, 'a_rule')],
            results,
        )

    def test_rule_does_not_take_current_rule(self):
        results = []

        class TestCheck:
            def __call__(self, target, cred, enforcer):
                results.append((target, cred, enforcer))
                return True

        # this is for the legacy path, and our type hints specifically don't
        # account for this path
        check = _checks.NotCheck(TestCheck())  # type: ignore

        self.assertFalse(check({}, {}, self.enforcer, current_rule='a_rule'))
        self.assertEqual(
            [({}, {}, self.enforcer)],
            results,
        )


class _BoolCheck(_checks.BaseCheck):
    def __init__(self, result):
        self.called = False
        self.result = result

    def __str__(self):
        return str(self.result)

    def __call__(self, target, creds, enforcer, current_rule=None):
        self.called = True
        return self.result


class AndCheckTestCase(base.PolicyBaseTestCase):
    def test_init(self):
        rule1 = CheckForTest('kind1', 'match1')
        rule2 = CheckForTest('kind2', 'match2')
        check = _checks.AndCheck([rule1, rule2])

        self.assertEqual([rule1, rule2], check.rules)

    def test_add_check(self):
        rule1 = CheckForTest('kind1', 'match1')
        rule2 = CheckForTest('kind2', 'match2')
        rule3 = CheckForTest('kind3', 'match3')
        check = _checks.AndCheck([rule1, rule2])
        check.add_check(rule3)

        self.assertEqual([rule1, rule2, rule3], check.rules)

    def test_str(self):
        rule1 = CheckForTest('kind1', 'match1')
        rule2 = CheckForTest('kind2', 'match2')
        check = _checks.AndCheck([rule1, rule2])

        self.assertEqual('(kind1:match1 and kind2:match2)', str(check))

    def test_call_all_false(self):
        rules = [_BoolCheck(False), _BoolCheck(False)]
        check = _checks.AndCheck(rules)

        self.assertFalse(check({}, {}, self.enforcer))
        self.assertTrue(rules[0].called)
        self.assertFalse(rules[1].called)

    def test_call_first_true(self):
        rules = [_BoolCheck(True), _BoolCheck(False)]
        check = _checks.AndCheck(rules)

        self.assertFalse(check({}, {}, self.enforcer))
        self.assertTrue(rules[0].called)
        self.assertTrue(rules[1].called)

    def test_call_second_true(self):
        rules = [_BoolCheck(False), _BoolCheck(True)]
        check = _checks.AndCheck(rules)

        self.assertFalse(check({}, {}, self.enforcer))
        self.assertTrue(rules[0].called)
        self.assertFalse(rules[1].called)

    def test_rule_takes_current_rule(self):
        results = []

        class TestCheck(_checks.Check):
            def __call__(self, target, cred, enforcer, current_rule=None):
                results.append((target, cred, enforcer, current_rule))
                return False

        check = _checks.AndCheck([TestCheck('kind', 'match')])

        self.assertFalse(check({}, {}, self.enforcer, current_rule='a_rule'))
        self.assertEqual(
            [({}, {}, self.enforcer, 'a_rule')],
            results,
        )

    def test_rule_does_not_take_current_rule(self):
        results = []

        class TestCheck:
            def __call__(self, target, cred, enforcer):
                results.append((target, cred, enforcer))
                return False

        # this is for the legacy path, and our type hints specifically don't
        # account for this path
        check = _checks.AndCheck([TestCheck()])  # type: ignore

        self.assertFalse(check({}, {}, self.enforcer, current_rule='a_rule'))
        self.assertEqual(
            [({}, {}, self.enforcer)],
            results,
        )


class OrCheckTestCase(base.PolicyBaseTestCase):
    def test_init(self):
        rule1 = CheckForTest('kind1', 'match1')
        rule2 = CheckForTest('kind2', 'match2')
        check = _checks.OrCheck([rule1, rule2])

        self.assertEqual([rule1, rule2], check.rules)

    def test_add_check(self):
        rule1 = CheckForTest('kind1', 'match1')
        rule2 = CheckForTest('kind2', 'match2')
        rule3 = CheckForTest('kind3', 'match3')
        check = _checks.OrCheck([rule1, rule2])
        check.add_check(rule3)

        self.assertEqual([rule1, rule2, rule3], check.rules)

    def test_pop_check(self):
        rule1 = CheckForTest('kind1', 'match1')
        rule2 = CheckForTest('kind2', 'match2')
        rule3 = CheckForTest('kind3', 'match3')
        check = _checks.OrCheck([rule1, rule2, rule3])
        rules, check1 = check.pop_check()

        self.assertEqual([rule1, rule2], check.rules)
        self.assertEqual(rule3, check1)

    def test_str(self):
        rule1 = CheckForTest('kind1', 'match1')
        rule2 = CheckForTest('kind2', 'match2')
        check = _checks.OrCheck([rule1, rule2])

        self.assertEqual('(kind1:match1 or kind2:match2)', str(check))

    def test_call_all_false(self):
        rules = [_BoolCheck(False), _BoolCheck(False)]
        check = _checks.OrCheck(rules)

        self.assertFalse(check({}, {}, self.enforcer))
        self.assertTrue(rules[0].called)
        self.assertTrue(rules[1].called)

    def test_call_first_true(self):
        rules = [_BoolCheck(True), _BoolCheck(False)]
        check = _checks.OrCheck(rules)

        self.assertTrue(check({}, {}, self.enforcer))
        self.assertTrue(rules[0].called)
        self.assertFalse(rules[1].called)

    def test_call_second_true(self):
        rules = [_BoolCheck(False), _BoolCheck(True)]
        check = _checks.OrCheck(rules)

        self.assertTrue(check({}, {}, self.enforcer))
        self.assertTrue(rules[0].called)
        self.assertTrue(rules[1].called)

    def test_rule_takes_current_rule(self):
        results = []

        class TestCheck(_checks.Check):
            def __call__(self, target, cred, enforcer, current_rule=None):
                results.append((target, cred, enforcer, current_rule))
                return False

        check = _checks.OrCheck([TestCheck('kind', 'match')])

        self.assertFalse(check({}, {}, self.enforcer, current_rule='a_rule'))
        self.assertEqual(
            [({}, {}, self.enforcer, 'a_rule')],
            results,
        )

    def test_rule_does_not_take_current_rule(self):
        results = []

        class TestCheck:
            def __call__(self, target, cred, enforcer):
                results.append((target, cred, enforcer))
                return False

        # this is a test for the legacy path: our type hints explicitly do not
        # handle this
        check = _checks.OrCheck([TestCheck()])  # type: ignore

        self.assertFalse(check({}, {}, self.enforcer, current_rule='a_rule'))
        self.assertEqual(
            [({}, {}, self.enforcer)],
            results,
        )



# Copyright (c) 2015 OpenStack Foundation.
# All Rights Reserved.

#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslotest import base as test_base

from oslo_policy import _checks
from oslo_policy import _parser
from oslo_policy.tests import base


class ParseCheckTestCase(test_base.BaseTestCase):
    def test_false(self):
        result = _parser._parse_check('!')

        self.assertIsInstance(result, _checks.FalseCheck)

    def test_true(self):
        result = _parser._parse_check('@')

        self.assertIsInstance(result, _checks.TrueCheck)

    @mock.patch.object(_parser, 'LOG')
    def test_bad_rule(self, mock_log):
        result = _parser._parse_check('foobar')

        self.assertIsInstance(result, _checks.FalseCheck)
        mock_log.exception.assert_called_once()

    @mock.patch.object(_checks, 'registered_checks', {})
    @mock.patch.object(_parser, 'LOG')
    def test_no_handler(self, mock_log):
        result = _parser._parse_check('no:handler')

        self.assertIsInstance(result, _checks.FalseCheck)
        mock_log.error.assert_called()

    @mock.patch.object(
        _checks,
        'registered_checks',
        {
            'spam': mock.Mock(return_value='spam_check'),
            None: mock.Mock(return_value='none_check'),
        },
    )
    def test_check(self):
        result = _parser._parse_check('spam:handler')

        self.assertEqual('spam_check', result)
        _checks.registered_checks['spam'].assert_called_once_with(  # type: ignore
            'spam', 'handler'
        )
        self.assertFalse(_checks.registered_checks[None].called)  # type: ignore

    @mock.patch.object(
        _checks,
        'registered_checks',
        {
            None: mock.Mock(return_value='none_check'),
        },
    )
    def test_check_default(self):
        result = _parser._parse_check('spam:handler')

        self.assertEqual('none_check', result)
        _checks.registered_checks[None].assert_called_once_with(  # type: ignore
            'spam', 'handler'
        )


class ParseListRuleTestCase(test_base.BaseTestCase):
    def test_empty(self):
        result = _parser._parse_list_rule([])

        self.assertIsInstance(result, _checks.TrueCheck)
        self.assertEqual('@', str(result))

    @mock.patch.object(_parser, '_parse_check', base.FakeCheck)
    def test_oneele_zeroele(self):
        result = _parser._parse_list_rule([[]])

        self.assertIsInstance(result, _checks.FalseCheck)
        self.assertEqual('!', str(result))

    @mock.patch.object(_parser, '_parse_check', base.FakeCheck)
    def test_oneele_bare(self):
        result = _parser._parse_list_rule(['rule'])

        self.assertIsInstance(result, base.FakeCheck)
        assert isinstance(result, base.FakeCheck)  # narrow type
        self.assertEqual('rule', result.result)
        self.assertEqual('rule', str(result))

    @mock.patch.object(_parser, '_parse_check', base.FakeCheck)
    def test_oneele_oneele(self):
        result = _parser._parse_list_rule([['rule']])

        self.assertIsInstance(result, base.FakeCheck)
        assert isinstance(result, base.FakeCheck)  # narrow type
        self.assertEqual('rule', result.result)
        self.assertEqual('rule', str(result))

    @mock.patch.object(_parser, '_parse_check', base.FakeCheck)
    def test_oneele_multi(self):
        result = _parser._parse_list_rule([['rule1', 'rule2']])

        self.assertIsInstance(result, _checks.AndCheck)
        assert isinstance(result, _checks.AndCheck)  # narrow type
        self.assertEqual(2, len(result.rules))
        for i, value in enumerate(['rule1', 'rule2']):
            rule = result.rules[i]
            self.assertIsInstance(rule, base.FakeCheck)
            assert isinstance(rule, base.FakeCheck)  # narrow type
            self.assertEqual(value, rule.result)
        self.assertEqual('(rule1 and rule2)', str(result))

    @mock.patch.object(_parser, '_parse_check', base.FakeCheck)
    def test_multi_oneele(self):
        result = _parser._parse_list_rule([['rule1'], ['rule2']])

        self.assertIsInstance(result, _checks.OrCheck)
        assert isinstance(result, _checks.OrCheck)  # narrow type
        self.assertEqual(2, len(result.rules))
        for i, value in enumerate(['rule1', 'rule2']):
            rule = result.rules[i]
            self.assertIsInstance(rule, base.FakeCheck)
            assert isinstance(rule, base.FakeCheck)  # narrow type
            self.assertEqual(value, rule.result)
        self.assertEqual('(rule1 or rule2)', str(result))

    @mock.patch.object(_parser, '_parse_check', base.FakeCheck)
    def test_multi_multi(self):
        result = _parser._parse_list_rule(
            [['rule1', 'rule2'], ['rule3', 'rule4']]
        )

        self.assertIsInstance(result, _checks.OrCheck)
        assert isinstance(result, _checks.OrCheck)  # narrow type
        self.assertEqual(2, len(result.rules))
        for i, values in enumerate([['rule1', 'rule2'], ['rule3', 'rule4']]):
            rule = result.rules[i]
            self.assertIsInstance(rule, _checks.AndCheck)
            assert isinstance(rule, _checks.AndCheck)  # narrow type
            self.assertEqual(2, len(rule.rules))
            for j, value in enumerate(values):
                sub_rule = rule.rules[j]
                assert isinstance(sub_rule, base.FakeCheck)  # narrow type
                self.assertIsInstance(sub_rule, base.FakeCheck)
                self.assertEqual(value, sub_rule.result)
        self.assertEqual(
            '((rule1 and rule2) or (rule3 and rule4))', str(result)
        )


class ParseTokenizeTestCase(test_base.BaseTestCase):
    def test_tokenize(self):
        exemplar = (
            '(( ( ((() And)) or ) (rule:context_is_admin) not)) '
            '\'a-string\' "another-string"'
        )

        result = list(_parser._parse_tokenize(exemplar))

        expected = [
            _parser.LeftParamToken('('),
            _parser.LeftParamToken('('),
            _parser.LeftParamToken('('),
            _parser.LeftParamToken('('),
            _parser.LeftParamToken('('),
            _parser.LeftParamToken('('),
            _parser.RightParamToken(')'),
            _parser.AndToken('And'),
            _parser.RightParamToken(')'),
            _parser.RightParamToken(')'),
            _parser.OrToken('or'),
            _parser.RightParamToken(')'),
            _parser.LeftParamToken('('),
            _parser.CheckToken(_checks.RuleCheck('rule', 'context_is_admin')),
            _parser.RightParamToken(')'),
            _parser.NotToken('not'),
            _parser.RightParamToken(')'),
            _parser.RightParamToken(')'),
            _parser.StringToken('a-string'),
            _parser.StringToken('another-string'),
        ]

        self.assertEqual(expected, result)


class ParseStateTestCase(test_base.BaseTestCase):
    def test_init(self):
        state = _parser.ParseState()

        self.assertEqual([], state.stack)

    def test_reduce_no_match(self):
        """Test that reduce() does nothing when no patterns match."""
        state = _parser.ParseState()
        # Create a token that won't match any reduction pattern
        unknown_token = _parser.StringToken('unknown_value')
        state.stack = [unknown_token]

        state.reduce()

        # Should remain unchanged
        self.assertEqual([unknown_token], state.stack)

    def test_reduce_parentheses(self):
        """Test reduction of parenthesized check."""
        state = _parser.ParseState()
        check = _checks.TrueCheck()
        state.stack = [
            _parser.LeftParamToken('('),
            _parser.CheckToken(check),
            _parser.RightParamToken(')'),
        ]

        state.reduce()

        self.assertEqual(1, len(state.stack))
        result_token = state.stack[0]
        self.assertIsInstance(result_token, _parser.CheckToken)
        self.assertEqual(check, result_token.value)

    def test_reduce_and_check(self):
        """Test reduction of AND expression with two checks."""
        state = _parser.ParseState()
        check1 = _checks.TrueCheck()
        check2 = _checks.FalseCheck()
        state.stack = [
            _parser.CheckToken(check1),
            _parser.AndToken('and'),
            _parser.CheckToken(check2),
        ]

        state.reduce()

        self.assertEqual(1, len(state.stack))
        result_token = state.stack[0]
        self.assertIsInstance(result_token, _parser.AndExprToken)
        result = result_token.value
        self.assertIsInstance(result, _checks.AndCheck)
        assert isinstance(result, _checks.AndCheck)  # narrow type
        self.assertEqual(2, len(result.rules))
        self.assertEqual(check1, result.rules[0])
        self.assertEqual(check2, result.rules[1])

    def test_reduce_or_check(self):
        """Test reduction of OR expression with two checks."""
        state = _parser.ParseState()
        check1 = _checks.TrueCheck()
        check2 = _checks.FalseCheck()
        state.stack = [
            _parser.CheckToken(check1),
            _parser.OrToken('or'),
            _parser.CheckToken(check2),
        ]

        state.reduce()

        self.assertEqual(1, len(state.stack))
        result_token = state.stack[0]
        self.assertIsInstance(result_token, _parser.OrExprToken)
        result = result_token.value
        self.assertIsInstance(result, _checks.OrCheck)
        assert isinstance(result, _checks.OrCheck)  # narrow type
        self.assertEqual(2, len(result.rules))
        self.assertEqual(check1, result.rules[0])
        self.assertEqual(check2, result.rules[1])

    def test_reduce_not_check(self):
        """Test reduction of NOT expression."""
        state = _parser.ParseState()
        check = _checks.TrueCheck()
        state.stack = [_parser.NotToken('not'), _parser.CheckToken(check)]

        state.reduce()

        self.assertEqual(1, len(state.stack))
        result_token = state.stack[0]
        self.assertIsInstance(result_token, _parser.CheckToken)
        result = result_token.value
        self.assertIsInstance(result, _checks.NotCheck)
        assert isinstance(result, _checks.NotCheck)  # narrow type
        self.assertEqual(check, result.rule)

    def test_shift(self):
        state = _parser.ParseState()

        with mock.patch.object(_parser.ParseState, 'reduce') as mock_reduce:
            token = _parser.StringToken('value')
            state.shift(token)

            self.assertEqual([token], state.stack)
            mock_reduce.assert_called_once_with()

    def test_result_empty(self):
        state = _parser.ParseState()

        self.assertRaises(ValueError, lambda: state.result)

    def test_result_unreduced(self):
        state = _parser.ParseState()
        state.stack = [
            _parser.StringToken('val1'),
            _parser.StringToken('val2'),
        ]

        self.assertRaises(ValueError, lambda: state.result)

    def test_result(self):
        state = _parser.ParseState()
        check = _checks.TrueCheck()
        state.stack = [_parser.CheckToken(check)]

        self.assertEqual(check, state.result)

    def test_result_invalid_value(self):
        """Test expression that does not resolve to check."""
        state = _parser.ParseState()
        # We can't reduce a StringToken down to a proper Check
        state.stack = [_parser.StringToken('not_a_check')]

        self.assertRaises(ValueError, lambda: state.result)

    def test_wrap_check_reduction(self):
        """Test parenthesized expression reduction via token shifting."""
        state = _parser.ParseState()
        check = _checks.TrueCheck()

        # Shift the tokens for a parenthesized expression
        state.shift(_parser.LeftParamToken('('))
        state.shift(_parser.CheckToken(check))
        state.shift(_parser.RightParamToken(')'))

        # Should reduce to a single check token
        self.assertEqual(1, len(state.stack))
        result_token = state.stack[0]
        self.assertIsInstance(result_token, _parser.CheckToken)
        self.assertEqual(check, result_token.value)
        self.assertEqual(check, state.result)

    def test_make_and_expr_reduction(self):
        """Test AND expression creation via token shifting."""
        state = _parser.ParseState()
        check1 = _checks.TrueCheck()
        check2 = _checks.FalseCheck()

        # Shift tokens for an AND expression
        state.shift(_parser.CheckToken(check1))
        state.shift(_parser.AndToken('and'))
        state.shift(_parser.CheckToken(check2))

        # Should reduce to a single and_expr token
        self.assertEqual(1, len(state.stack))
        result_token = state.stack[0]
        self.assertIsInstance(result_token, _parser.AndExprToken)
        result = result_token.value
        self.assertIsInstance(result, _checks.AndCheck)
        assert isinstance(result, _checks.AndCheck)  # narrow type
        self.assertEqual(2, len(result.rules))
        self.assertEqual(check1, result.rules[0])
        self.assertEqual(check2, result.rules[1])

    def test_extend_and_expr_reduction(self):
        """Test AND expression extension via token shifting."""
        state = _parser.ParseState()
        check1 = _checks.TrueCheck()
        check2 = _checks.FalseCheck()
        check3 = _checks.TrueCheck()

        # First create an and_expr
        state.shift(_parser.CheckToken(check1))
        state.shift(_parser.AndToken('and'))
        state.shift(_parser.CheckToken(check2))

        # Should have an and_expr now
        self.assertEqual(1, len(state.stack))
        self.assertIsInstance(state.stack[0], _parser.AndExprToken)

        # Now extend it with another check
        state.shift(_parser.AndToken('and'))
        state.shift(_parser.CheckToken(check3))

        # Should still be a single and_expr token but with 3 checks
        self.assertEqual(1, len(state.stack))
        result_token = state.stack[0]
        self.assertIsInstance(result_token, _parser.AndExprToken)
        result = result_token.value
        self.assertIsInstance(result, _checks.AndCheck)
        assert isinstance(result, _checks.AndCheck)  # narrow type
        self.assertEqual(3, len(result.rules))
        self.assertEqual(check1, result.rules[0])
        self.assertEqual(check2, result.rules[1])
        self.assertEqual(check3, result.rules[2])

    def test_make_or_expr_reduction(self):
        """Test OR expression creation via token shifting."""
        state = _parser.ParseState()
        check1 = _checks.TrueCheck()
        check2 = _checks.FalseCheck()

        # Shift tokens for an OR expression
        state.shift(_parser.CheckToken(check1))
        state.shift(_parser.OrToken('or'))
        state.shift(_parser.CheckToken(check2))

        # Should reduce to a single or_expr token
        self.assertEqual(1, len(state.stack))
        result_token = state.stack[0]
        self.assertIsInstance(result_token, _parser.OrExprToken)
        result = result_token.value
        self.assertIsInstance(result, _checks.OrCheck)
        assert isinstance(result, _checks.OrCheck)  # narrow type
        self.assertEqual(2, len(result.rules))
        self.assertEqual(check1, result.rules[0])
        self.assertEqual(check2, result.rules[1])

    def test_extend_or_expr_reduction(self):
        """Test OR expression extension via token shifting."""
        state = _parser.ParseState()
        check1 = _checks.TrueCheck()
        check2 = _checks.FalseCheck()
        check3 = _checks.TrueCheck()

        # First create an or_expr
        state.shift(_parser.CheckToken(check1))
        state.shift(_parser.OrToken('or'))
        state.shift(_parser.CheckToken(check2))

        # Should have an or_expr now
        self.assertEqual(1, len(state.stack))
        self.assertIsInstance(state.stack[0], _parser.OrExprToken)

        # Now extend it with another check
        state.shift(_parser.OrToken('or'))
        state.shift(_parser.CheckToken(check3))

        # Should still be a single or_expr token but with 3 checks
        self.assertEqual(1, len(state.stack))
        result_token = state.stack[0]
        self.assertIsInstance(result_token, _parser.OrExprToken)
        result = result_token.value
        self.assertIsInstance(result, _checks.OrCheck)
        assert isinstance(result, _checks.OrCheck)  # narrow type
        self.assertEqual(3, len(result.rules))
        self.assertEqual(check1, result.rules[0])
        self.assertEqual(check2, result.rules[1])
        self.assertEqual(check3, result.rules[2])

    def test_make_not_expr_reduction(self):
        """Test NOT expression creation via token shifting."""
        state = _parser.ParseState()
        check = _checks.TrueCheck()

        # Shift tokens for a NOT expression
        state.shift(_parser.NotToken('not'))
        state.shift(_parser.CheckToken(check))

        # Should reduce to a single check token with NotCheck
        self.assertEqual(1, len(state.stack))
        result_token = state.stack[0]
        self.assertIsInstance(result_token, _parser.CheckToken)
        result = result_token.value
        self.assertIsInstance(result, _checks.NotCheck)
        assert isinstance(result, _checks.NotCheck)  # narrow type
        self.assertEqual(check, result.rule)

    def test_and_expr_with_or_reduction(self):
        """Test OR expression with AND expression via token shifting."""
        state = _parser.ParseState()
        check1 = _checks.TrueCheck()
        check2 = _checks.FalseCheck()
        check3 = _checks.TrueCheck()

        # Create and_expr first
        state.shift(_parser.CheckToken(check1))
        state.shift(_parser.AndToken('and'))
        state.shift(_parser.CheckToken(check2))

        # Should have an and_expr
        self.assertEqual(1, len(state.stack))
        self.assertIsInstance(state.stack[0], _parser.AndExprToken)

        # Now create an OR with it
        state.shift(_parser.OrToken('or'))
        state.shift(_parser.CheckToken(check3))

        # Should reduce to a single or_expr token
        self.assertEqual(1, len(state.stack))
        result_token = state.stack[0]
        self.assertIsInstance(result_token, _parser.OrExprToken)
        result = result_token.value
        self.assertIsInstance(result, _checks.OrCheck)
        assert isinstance(result, _checks.OrCheck)  # narrow type
        self.assertEqual(2, len(result.rules))
        # First rule should be the AndCheck
        self.assertIsInstance(result.rules[0], _checks.AndCheck)
        # Second rule should be the simple check
        self.assertEqual(check3, result.rules[1])

    def test_or_expr_with_and_reduction(self):
        """Test OR expression with AND precedence via token shifting.

        This tests the case 'A or B and C' which should be parsed as 'A or (B
        and C)'.
        """
        state = _parser.ParseState()
        check1 = _checks.TrueCheck()
        check2 = _checks.FalseCheck()
        check3 = _checks.TrueCheck()

        # Create or_expr first
        state.shift(_parser.CheckToken(check1))
        state.shift(_parser.OrToken('or'))
        state.shift(_parser.CheckToken(check2))

        # Should have an or_expr
        self.assertEqual(1, len(state.stack))
        result_token = state.stack[0]
        self.assertIsInstance(result_token, _parser.OrExprToken)
        initial_or = result_token.value
        self.assertIsInstance(initial_or, _checks.OrCheck)

        # Now add AND with higher precedence which should modify the last check
        state.shift(_parser.AndToken('and'))
        state.shift(_parser.CheckToken(check3))

        # Should still be a single or_expr token
        self.assertEqual(1, len(state.stack))
        final_token = state.stack[0]
        self.assertIsInstance(final_token, _parser.OrExprToken)
        result = final_token.value
        self.assertIsInstance(result, _checks.OrCheck)
        assert isinstance(result, _checks.OrCheck)  # narrow type
        self.assertEqual(2, len(result.rules))
        # First rule should be the simple check1
        self.assertEqual(check1, result.rules[0])
        # Second rule should be an AndCheck of check2 and check3
        self.assertIsInstance(result.rules[1], _checks.AndCheck)
        and_rule = result.rules[1]
        assert isinstance(and_rule, _checks.AndCheck)  # narrow type
        self.assertEqual(2, len(and_rule.rules))
        self.assertEqual(check2, and_rule.rules[0])
        self.assertEqual(check3, and_rule.rules[1])


class ParseTextRuleTestCase(test_base.BaseTestCase):
    def test_empty(self):
        result = _parser._parse_text_rule('')

        self.assertIsInstance(result, _checks.TrueCheck)

    @mock.patch.object(_parser.ParseState, 'shift')
    @mock.patch.object(_parser.ParseState, 'result', 'result')
    def test_shifts(self, mock_shift):
        # Create mock tokens
        token1 = _parser.StringToken('val1')
        token2 = _parser.StringToken('val2')

        with mock.patch.object(
            _parser, '_parse_tokenize', return_value=[token1, token2]
        ):
            result = _parser._parse_text_rule('test rule')

            self.assertEqual('result', result)
            mock_shift.assert_has_calls([mock.call(token1), mock.call(token2)])

    @mock.patch.object(_parser, 'LOG', new=mock.Mock())
    @mock.patch.object(_parser, '_parse_tokenize', return_value=[])
    def test_fail(self, mock_parse_tokenize):
        result = _parser._parse_text_rule('test rule')

        self.assertIsInstance(result, _checks.FalseCheck)
        mock_parse_tokenize.assert_called_once_with('test rule')

    def test_A_or_B_or_C(self):
        result = _parser._parse_text_rule('@ or ! or @')
        self.assertEqual('(@ or ! or @)', str(result))

    def test_A_or_B_and_C(self):
        result = _parser._parse_text_rule('@ or ! and @')
        self.assertEqual('(@ or (! and @))', str(result))

    def test_A_and_B_or_C(self):
        result = _parser._parse_text_rule('@ and ! or @')
        self.assertEqual('((@ and !) or @)', str(result))

    def test_A_and_B_and_C(self):
        result = _parser._parse_text_rule('@ and ! and @')
        self.assertEqual('(@ and ! and @)', str(result))

    def test_A_or_B_or_C_or_D(self):
        result = _parser._parse_text_rule('@ or ! or @ or !')
        self.assertEqual('(@ or ! or @ or !)', str(result))

    def test_A_or_B_or_C_and_D(self):
        result = _parser._parse_text_rule('@ or ! or @ and !')
        self.assertEqual('(@ or ! or (@ and !))', str(result))

    def test_A_or_B_and_C_or_D(self):
        result = _parser._parse_text_rule('@ or ! and @ or !')
        self.assertEqual('(@ or (! and @) or !)', str(result))

    def test_A_or_B_and_C_and_D(self):
        result = _parser._parse_text_rule('@ or ! and @ and !')
        self.assertEqual('(@ or (! and @ and !))', str(result))

    def test_A_and_B_or_C_or_D(self):
        result = _parser._parse_text_rule('@ and ! or @ or !')
        self.assertEqual('((@ and !) or @ or !)', str(result))

    def test_A_and_B_or_C_and_D(self):
        result = _parser._parse_text_rule('@ and ! or @ and !')
        self.assertEqual('((@ and !) or (@ and !))', str(result))

    def test_A_and_B_and_C_or_D(self):
        result = _parser._parse_text_rule('@ and ! and @ or !')
        self.assertEqual('((@ and ! and @) or !)', str(result))

    def test_A_and_B_and_C_and_D(self):
        result = _parser._parse_text_rule('@ and ! and @ and !')
        self.assertEqual('(@ and ! and @ and !)', str(result))

    def test_A_and_B_or_C_with_not_1(self):
        result = _parser._parse_text_rule('not @ and ! or @')
        self.assertEqual('((not @ and !) or @)', str(result))

    def test_A_and_B_or_C_with_not_2(self):
        result = _parser._parse_text_rule('@ and not ! or @')
        self.assertEqual('((@ and not !) or @)', str(result))

    def test_A_and_B_or_C_with_not_3(self):
        result = _parser._parse_text_rule('@ and ! or not @')
        self.assertEqual('((@ and !) or not @)', str(result))

    def test_A_and_B_or_C_with_group_1(self):
        for expression in [
            '( @ ) and ! or @',
            '@ and ( ! ) or @',
            '@ and ! or ( @ )',
            '( @ ) and ! or ( @ )',
            '@ and ( ! ) or ( @ )',
            '( @ ) and ( ! ) or ( @ )',
            '( @ and ! ) or @',
            '( ( @ ) and ! ) or @',
            '( @ and ( ! ) ) or @',
            '( ( @ and ! ) ) or @',
            '( @ and ! or @ )',
        ]:
            result = _parser._parse_text_rule(expression)
            self.assertEqual('((@ and !) or @)', str(result))

    def test_A_and_B_or_C_with_group_2(self):
        result = _parser._parse_text_rule('@ and ( ! or @ )')
        self.assertEqual('(@ and (! or @))', str(result))

    def test_A_and_B_or_C_with_group_and_not_1(self):
        for expression in [
            'not ( @ ) and ! or @',
            'not @ and ( ! ) or @',
            'not @ and ! or ( @ )',
            '( not @ ) and ! or @',
            '( not @ and ! ) or @',
            '( not @ and ! or @ )',
        ]:
            result = _parser._parse_text_rule(expression)
            self.assertEqual('((not @ and !) or @)', str(result))

    def test_A_and_B_or_C_with_group_and_not_2(self):
        result = _parser._parse_text_rule('not @ and ( ! or @ )')
        self.assertEqual('(not @ and (! or @))', str(result))

    def test_A_and_B_or_C_with_group_and_not_3(self):
        result = _parser._parse_text_rule('not ( @ and ! or @ )')
        self.assertEqual('not ((@ and !) or @)', str(result))

    def test_A_and_B_or_C_with_group_and_not_4(self):
        for expression in [
            '( @ ) and not ! or @',
            '@ and ( not ! ) or @',
            '@ and not ( ! ) or @',
            '@ and not ! or ( @ )',
            '( @ and not ! ) or @',
            '( @ and not ! or @ )',
        ]:
            result = _parser._parse_text_rule(expression)
            self.assertEqual('((@ and not !) or @)', str(result))

    def test_A_and_B_or_C_with_group_and_not_5(self):
        result = _parser._parse_text_rule('@ and ( not ! or @ )')
        self.assertEqual('(@ and (not ! or @))', str(result))

    def test_A_and_B_or_C_with_group_and_not_6(self):
        result = _parser._parse_text_rule('@ and not ( ! or @ )')
        self.assertEqual('(@ and not (! or @))', str(result))

    def test_A_and_B_or_C_with_group_and_not_7(self):
        for expression in [
            '( @ ) and ! or not @',
            '@ and ( ! ) or not @',
            '@ and ! or not ( @ )',
            '@ and ! or ( not @ )',
            '( @ and ! ) or not @',
            '( @ and ! or not @ )',
        ]:
            result = _parser._parse_text_rule(expression)
            self.assertEqual('((@ and !) or not @)', str(result))

    def test_A_and_B_or_C_with_group_and_not_8(self):
        result = _parser._parse_text_rule('@ and ( ! or not @ )')
        self.assertEqual('(@ and (! or not @))', str(result))


class ParseRuleTestCase(test_base.BaseTestCase):
    @mock.patch.object(_parser, '_parse_text_rule', return_value='text rule')
    @mock.patch.object(_parser, '_parse_list_rule', return_value='list rule')
    def test_parse_rule_string(
        self, mock_parse_list_rule, mock_parse_text_rule
    ):
        result = _parser.parse_rule('a string')

        self.assertEqual('text rule', result)
        self.assertFalse(mock_parse_list_rule.called)
        mock_parse_text_rule.assert_called_once_with('a string')

    @mock.patch.object(_parser, '_parse_text_rule', return_value='text rule')
    @mock.patch.object(_parser, '_parse_list_rule', return_value='list rule')
    def test_parse_rule_list(self, mock_parse_list_rule, mock_parse_text_rule):
        result = _parser.parse_rule([['a'], ['list']])

        self.assertEqual('list rule', result)
        self.assertFalse(mock_parse_text_rule.called)
        mock_parse_list_rule.assert_called_once_with([['a'], ['list']])



# Copyright 2017 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import textwrap

from oslotest import base

from oslo_policy import policy
from oslo_policy import sphinxext


class IndentTest(base.BaseTestCase):
    def test_indent(self):
        result = sphinxext._indent('foo\nbar')
        self.assertEqual('    foo\n    bar', result)

        result = sphinxext._indent('')
        self.assertEqual('', result)

        result = sphinxext._indent('\n')
        self.assertEqual('\n', result)

        result = sphinxext._indent('test\ntesting\n\nafter blank')
        self.assertEqual('    test\n    testing\n\n    after blank', result)

        result = sphinxext._indent('\tfoo\nbar')
        self.assertEqual('    \tfoo\n    bar', result)

        result = sphinxext._indent('    foo\nbar')
        self.assertEqual('        foo\n    bar', result)

        result = sphinxext._indent('foo\n    bar')
        self.assertEqual('    foo\n        bar', result)

        result = sphinxext._indent('foo\n\n    bar')
        self.assertEqual('    foo\n\n        bar', result)

        self.assertRaises(AttributeError, sphinxext._indent, None)


class FormatPolicyTest(base.BaseTestCase):
    def test_minimal(self):
        results = '\n'.join(
            list(
                sphinxext._format_policy_section(
                    'foo', [policy.RuleDefault('rule_a', '@')]
                )
            )
        )

        self.assertEqual(
            textwrap.dedent("""
        foo
        ===

        ``rule_a``
            :Default: ``@``

            (no description provided)
        """).lstrip(),
            results,
        )

    def test_with_description(self):
        results = '\n'.join(
            list(
                sphinxext._format_policy_section(
                    'foo',
                    [policy.RuleDefault('rule_a', '@', 'My sample rule')],
                )
            )
        )

        self.assertEqual(
            textwrap.dedent("""
        foo
        ===

        ``rule_a``
            :Default: ``@``

            My sample rule
        """).lstrip(),
            results,
        )

    def test_with_operations(self):
        results = '\n'.join(
            list(
                sphinxext._format_policy_section(
                    'foo',
                    [
                        policy.DocumentedRuleDefault(
                            'rule_a',
                            '@',
                            'My sample rule',
                            [
                                {'method': 'GET', 'path': '/foo'},
                                {'method': 'POST', 'path': '/some'},
                            ],
                        )
                    ],
                )
            )
        )

        self.assertEqual(
            textwrap.dedent("""
        foo
        ===

        ``rule_a``
            :Default: ``@``
            :Operations:
                - **GET** ``/foo``
                - **POST** ``/some``

            My sample rule
        """).lstrip(),
            results,
        )

    def test_with_scope_types(self):
        operations: list[policy.Operation] = [
            {'method': 'GET', 'path': '/foo'},
            {'method': 'POST', 'path': '/some'},
        ]
        scope_types = ['bar']
        rule = policy.DocumentedRuleDefault(
            'rule_a',
            '@',
            'My sample rule',
            operations,
            scope_types=scope_types,
        )

        results = '\n'.join(
            list(sphinxext._format_policy_section('foo', [rule]))
        )

        self.assertEqual(
            textwrap.dedent("""
        foo
        ===

        ``rule_a``
            :Default: ``@``
            :Operations:
                - **GET** ``/foo``
                - **POST** ``/some``
            :Scope Types:
                - **bar**

            My sample rule
        """).lstrip(),
            results,
        )



#
# Copyright (c) 2015 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
import copy
import os
from typing import Any, TYPE_CHECKING

import requests
from requests.exceptions import Timeout

from oslo_policy import _checks
from oslo_policy._i18n import _
from oslo_serialization import jsonutils

if TYPE_CHECKING:
    from oslo_policy.policy import Enforcer


class HttpCheck(_checks.Check):
    """Check ``http:`` rules by calling to a remote server.

    This example implementation simply verifies that the response
    is exactly ``True``.
    """

    def __call__(
        self,
        target: _checks.TargetT,
        creds: _checks.CredsT,
        enforcer: 'Enforcer',
        current_rule: str | None = None,
    ) -> bool:
        timeout = enforcer.conf.oslo_policy.remote_timeout

        url = ('http:' + self.match) % target
        data, json = self._construct_payload(
            creds, current_rule, enforcer, target
        )
        try:
            with contextlib.closing(
                requests.post(url, json=json, data=data, timeout=timeout)
            ) as r:
                return bool(r.text.lstrip('"').rstrip('"') == 'True')
        except Timeout:
            raise RuntimeError('Timeout in REST API call')

    @staticmethod
    def _construct_payload(
        creds: _checks.CredsT,
        current_rule: str | None,
        enforcer: 'Enforcer',
        target: _checks.TargetT,
    ) -> tuple[dict[str, Any], None] | tuple[None, dict[str, Any]]:
        # Convert instances of object() in target temporarily to
        # empty dict to avoid circular reference detection
        # errors in jsonutils.dumps().
        temp_target = copy.deepcopy(target)
        for key in target.keys():
            element = target.get(key)
            if type(element) is object:
                temp_target[key] = {}  # type: ignore
        if (
            enforcer.conf.oslo_policy.remote_content_type
            == 'application/x-www-form-urlencoded'
        ):
            data = {
                'rule': jsonutils.dumps(current_rule),
                'target': jsonutils.dumps(temp_target),
                'credentials': jsonutils.dumps(creds),
            }
            json = None
            return data, json
        else:
            data = None
            json = {
                'rule': current_rule,
                'target': temp_target,
                'credentials': creds,
            }
            return data, json


class HttpsCheck(HttpCheck):
    """Check ``https:`` rules by calling to a remote server.

    This example implementation simply verifies that the response
    is exactly ``True``.
    """

    def __call__(
        self,
        target: _checks.TargetT,
        creds: _checks.CredsT,
        enforcer: 'Enforcer',
        current_rule: str | None = None,
    ) -> bool:
        url = ('https:' + self.match) % target

        cert_file = enforcer.conf.oslo_policy.remote_ssl_client_crt_file
        key_file = enforcer.conf.oslo_policy.remote_ssl_client_key_file
        ca_crt_file = enforcer.conf.oslo_policy.remote_ssl_ca_crt_file
        verify_server = enforcer.conf.oslo_policy.remote_ssl_verify_server_crt
        timeout = enforcer.conf.oslo_policy.remote_timeout

        if cert_file:
            if not os.path.exists(cert_file):
                raise RuntimeError(
                    _('Unable to find ssl cert_file  : %s') % cert_file
                )
            if not os.access(cert_file, os.R_OK):
                raise RuntimeError(
                    _('Unable to access ssl cert_file  : %s') % cert_file
                )
        if key_file:
            if not os.path.exists(key_file):
                raise RuntimeError(
                    _('Unable to find ssl key_file : %s') % key_file
                )
            if not os.access(key_file, os.R_OK):
                raise RuntimeError(
                    _('Unable to access ssl key_file  : %s') % key_file
                )
        cert = (cert_file, key_file)
        if verify_server:
            if ca_crt_file:
                if not os.path.exists(ca_crt_file):
                    raise RuntimeError(
                        _('Unable to find ca cert_file  : %s') % ca_crt_file
                    )
                verify_server = ca_crt_file

        data, json = self._construct_payload(
            creds, current_rule, enforcer, target
        )
        try:
            with contextlib.closing(
                requests.post(
                    url,
                    json=json,
                    data=data,
                    cert=cert,
                    verify=verify_server,
                    timeout=timeout,
                )
            ) as r:
                return bool(r.text.lstrip('"').rstrip('"') == 'True')
        except Timeout:
            raise RuntimeError('Timeout in REST API call')



#
# Copyright (c) 2015 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import re
from collections.abc import Generator, Sequence
from typing import Any, Generic, TypeAlias, TypeVar

from oslo_policy import _checks

LOG = logging.getLogger(__name__)

T = TypeVar('T')
TokenT: TypeAlias = str
ValueT: TypeAlias = str | _checks.BaseCheck


class Token(Generic[T]):
    def __init__(self, value: T) -> None:
        self.value = value

    def __repr__(self) -> str:
        return f'<{self.__class__.__qualname__}({self.value})'

    def __eq__(self, other: Any) -> bool:
        return type(self) is type(other) and self.value == other.value


# fmt: off
class CheckToken(Token[_checks.BaseCheck]): ...
class AndExprToken(Token[_checks.AndCheck]): ...
class OrExprToken(Token[_checks.OrCheck]): ...
class AndToken(Token[str]): ...
class OrToken(Token[str]): ...
class NotToken(Token[str]): ...
class LeftParamToken(Token[str]): ...
class RightParamToken(Token[str]): ...
class StringToken(Token[str]): ...
# fmt: on


class ParseState:
    """Implement the core of parsing the policy language.

    Uses a greedy reduction algorithm to reduce a sequence of tokens into
    a single terminal, the value of which will be the root of the
    :class:`Check` tree.

    .. note::

        Error reporting is rather lacking.  The best we can get with this
        parser formulation is an overall "parse failed" error. Fortunately, the
        policy language is simple enough that this shouldn't be that big a
        problem.
    """

    def __init__(self) -> None:
        """Initialize the ParseState."""
        self.stack: list[Token[Any]] = []

    def reduce(self) -> None:
        """Perform a greedy reduction of the token stream.

        Uses pattern matching to efficiently find and apply reduction rules.
        If a rule matches, it will be executed, then the method will be called
        recursively to search for any more possible reductions.
        """
        # Try 3-token patterns first
        if len(self.stack) >= 3:
            token_a, token_b, token_c = self.stack[-3:]
            match [token_a, token_b, token_c]:
                # Parenthesized expressions
                case (
                    [LeftParamToken(), CheckToken(), RightParamToken()]
                    | [LeftParamToken(), AndExprToken(), RightParamToken()]
                    | [LeftParamToken(), OrExprToken(), RightParamToken()]
                ):
                    # Turn parenthesized expressions into a 'check' token
                    self.stack[-3:] = [CheckToken(token_b.value)]
                    return self.reduce()

                # AND expressions
                case [CheckToken(), AndToken(), CheckToken()]:
                    # Create an 'and_expr' - join two checks by the 'and'
                    # operator
                    check1, check2 = token_a.value, token_c.value
                    and_check = _checks.AndCheck([check1, check2])
                    self.stack[-3:] = [AndExprToken(and_check)]
                    return self.reduce()

                case [AndExprToken(), AndToken(), CheckToken()]:
                    # Extend an 'and_expr' by adding one more check
                    and_expr, check = token_a.value, token_c.value
                    extended_and = and_expr.add_check(check)
                    self.stack[-3:] = [AndExprToken(extended_and)]
                    return self.reduce()

                case [OrExprToken(), AndToken(), CheckToken()]:
                    # Modify the case 'A or B and C'
                    or_expr, check = token_a.value, token_c.value
                    or_expr_check, check1 = or_expr.pop_check()
                    if isinstance(check1, _checks.AndCheck):
                        and_expr = check1.add_check(check)
                    else:
                        and_expr = _checks.AndCheck([check1, check])
                    result_or = or_expr_check.add_check(and_expr)
                    self.stack[-3:] = [OrExprToken(result_or)]
                    return self.reduce()

                # OR expressions
                case [CheckToken(), OrToken(), CheckToken()] | [
                    AndExprToken(),
                    OrToken(),
                    CheckToken(),
                ]:
                    # Create an 'or_expr' - join two checks by the 'or'
                    # operator
                    check1, check2 = token_a.value, token_c.value
                    or_check = _checks.OrCheck([check1, check2])
                    self.stack[-3:] = [OrExprToken(or_check)]
                    return self.reduce()

                case [OrExprToken(), OrToken(), CheckToken()]:
                    # Extend an 'or_expr' by adding one more check
                    or_expr, check = token_a.value, token_c.value
                    extended_or = or_expr.add_check(check)
                    self.stack[-3:] = [OrExprToken(extended_or)]
                    return self.reduce()

        # Try 2-token patterns
        if len(self.stack) >= 2:
            token_a, token_b = self.stack[-2:]
            match self.stack[-2:]:
                # NOT expressions
                case [NotToken(), CheckToken()]:
                    # Invert the result of another check
                    check = token_b.value
                    not_check = _checks.NotCheck(check)
                    self.stack[-2:] = [CheckToken(not_check)]
                    return self.reduce()

    def shift(self, token: Token[Any]) -> None:
        """Adds one more token to the state.

        Calls :meth:`reduce`.
        """
        self.stack.append(token)

        # Do a greedy reduce...
        self.reduce()

    @property
    def result(self) -> _checks.BaseCheck:
        """Obtain the final result of the parse.

        :raises ValueError: If the parse failed to reduce to a single result.
        """
        if len(self.stack) != 1:
            raise ValueError('Could not parse rule')

        value = self.stack[0].value
        if not isinstance(value, _checks.BaseCheck):
            # we should never get here since we should have reduced out any
            # string tokens
            raise ValueError('Could not parse rule')

        return value


def _parse_check(rule: str) -> _checks.BaseCheck:
    """Parse a single base check rule into an appropriate Check object."""
    # Handle the special checks
    if rule == '!':
        return _checks.FalseCheck()
    elif rule == '@':
        return _checks.TrueCheck()

    try:
        kind, match = rule.split(':', 1)
    except Exception:
        LOG.exception('Failed to understand rule %s', rule)
        # If the rule is invalid, we'll fail closed
        return _checks.FalseCheck()

    # Find what implements the check
    extension_checks = _checks.get_extensions()
    if kind in extension_checks:
        return extension_checks[kind](kind, match)
    elif kind in _checks.registered_checks:
        return _checks.registered_checks[kind](kind, match)
    elif None in _checks.registered_checks:
        return _checks.registered_checks[None](kind, match)
    else:
        LOG.error('No handler for matches of kind %s', kind)
        return _checks.FalseCheck()


def _parse_list_rule(rule: Sequence[str | Sequence[str]]) -> _checks.BaseCheck:
    """Translates the old list-of-lists syntax into a tree of Check objects.

    Provided for backwards compatibility.
    """
    # Empty rule defaults to True
    if not rule:
        return _checks.TrueCheck()

    # Outer list is joined by "or"; inner list by "and"
    or_list: list[_checks.BaseCheck] = []
    for inner_rule in rule:
        # Skip empty inner lists
        if not inner_rule:
            continue

        # Handle bare strings
        if isinstance(inner_rule, str):
            inner_rule = [inner_rule]

        # Parse the inner rules into Check objects
        and_list = [_parse_check(r) for r in inner_rule]

        # Append the appropriate check to the or_list
        if len(and_list) == 1:
            or_list.append(and_list[0])
        else:
            or_list.append(_checks.AndCheck(and_list))

    # If we have only one check, omit the "or"
    if not or_list:
        return _checks.FalseCheck()
    elif len(or_list) == 1:
        return or_list[0]

    return _checks.OrCheck(or_list)


# Used for tokenizing the policy language
_tokenize_re = re.compile(r'\s+')


def _parse_tokenize(rule: str) -> Generator[Token[Any], None, None]:
    """Tokenizer for the policy language.

    Most of the single-character tokens are specified in the
    ``_tokenize_re``; however, parentheses need to be handled specially,
    because they can appear inside a check string.  Thankfully, those
    parentheses that appear inside a check string can never occur at
    the very beginning or end (``"%(variable)s"`` is the correct syntax).
    """
    for tok in _tokenize_re.split(rule):
        # Skip empty tokens
        if not tok or tok.isspace():
            continue

        # Handle leading parens on the token
        clean = tok.lstrip('(')
        for i in range(len(tok) - len(clean)):
            # yes, the value argument is redundant but meh
            yield LeftParamToken('(')

        # If it was only parentheses, continue
        if not clean:
            continue
        else:
            tok = clean

        # Handle trailing parens on the token
        clean = tok.rstrip(')')
        trail = len(tok) - len(clean)

        # Yield the cleaned token
        lowered = clean.lower()
        if lowered in ('and', 'or', 'not'):
            match lowered:
                case 'and':
                    yield AndToken(clean)
                case 'or':
                    yield OrToken(clean)
                case 'not':
                    yield NotToken(clean)
        elif clean:
            # Not a special token, but not composed solely of ')'
            if len(tok) >= 2 and (
                (tok[0], tok[-1]) in [('"', '"'), ("'", "'")]
            ):
                # It's a quoted string: drop the quotes
                yield StringToken(tok[1:-1])
            else:
                yield CheckToken(_parse_check(clean))

        # Yield the trailing parens
        for i in range(trail):
            yield RightParamToken(')')


def _parse_text_rule(rule: str) -> _checks.BaseCheck:
    """Parses policy to the tree.

    Translates a policy written in the policy language into a tree of
    Check objects.
    """
    # Empty rule means always accept
    if not rule:
        return _checks.TrueCheck()

    # Parse the token stream
    state = ParseState()
    for token in _parse_tokenize(rule):
        state.shift(token)

    try:
        return state.result
    except ValueError:
        # Couldn't parse the rule
        LOG.exception('Failed to understand rule %s', rule)

        # Fail closed
        return _checks.FalseCheck()


def parse_rule(rule: str | Sequence[str | Sequence[str]]) -> _checks.BaseCheck:
    """Parses a policy rule into a tree of :class:`.Check` objects."""
    # If the rule is a string, it's in the policy language
    if isinstance(rule, str):
        return _parse_text_rule(rule)
    return _parse_list_rule(rule)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslotest import base

from oslo_policy import sphinxpolicygen


class SingleSampleGenerationTest(base.BaseTestCase):
    @mock.patch('os.path.isdir')
    @mock.patch('os.path.isfile')
    @mock.patch('oslo_policy.generator.generate_sample')
    def test_sample_gen_with_single_config_file(self, sample, isfile, isdir):
        isfile.side_effect = [False, True]
        isdir.return_value = True

        config = mock.Mock(
            policy_generator_config_file='nova.conf',
            sample_policy_basename='nova',
            exclude_deprecated=False,
        )
        app = mock.Mock(srcdir='/opt/nova', config=config)
        sphinxpolicygen.generate_sample(app)

        sample.assert_called_once_with(
            args=[
                '--config-file',
                '/opt/nova/nova.conf',
                '--output-file',
                '/opt/nova/nova.policy.yaml.sample',
            ],
            conf=mock.ANY,
        )

    @mock.patch('os.path.isdir')
    @mock.patch('os.path.isfile')
    @mock.patch('oslo_policy.generator.generate_sample')
    def test_sample_gen_with_single_config_file_no_base(
        self, sample, isfile, isdir
    ):
        isfile.side_effect = [False, True]
        isdir.return_value = True

        config = mock.Mock(
            policy_generator_config_file='nova.conf',
            sample_policy_basename=None,
            exclude_deprecated=True,
        )
        app = mock.Mock(srcdir='/opt/nova', config=config)
        sphinxpolicygen.generate_sample(app)

        sample.assert_called_once_with(
            args=[
                '--config-file',
                '/opt/nova/nova.conf',
                '--output-file',
                '/opt/nova/sample.policy.yaml',
                '--exclude-deprecated',
            ],
            conf=mock.ANY,
        )

    @mock.patch('os.path.isdir')
    @mock.patch('os.path.isfile')
    @mock.patch('oslo_policy.generator.generate_sample')
    def test_sample_gen_with_multiple_config_files(
        self, sample, isfile, isdir
    ):
        # Tests the scenario that policy_generator_config_file is a list
        # of two-item tuples of the config file name and policy basename.
        isfile.side_effect = [False, True] * 2
        isdir.return_value = True

        config = mock.Mock(
            policy_generator_config_file=[
                ('nova.conf', 'nova'),
                ('placement.conf', 'placement'),
            ],
            exclude_deprecated=False,
        )
        app = mock.Mock(srcdir='/opt/nova', config=config)
        sphinxpolicygen.generate_sample(app)

        sample.assert_has_calls(
            [
                mock.call(
                    args=[
                        '--config-file',
                        '/opt/nova/nova.conf',
                        '--output-file',
                        '/opt/nova/nova.policy.yaml.sample',
                    ],
                    conf=mock.ANY,
                ),
                mock.call(
                    args=[
                        '--config-file',
                        '/opt/nova/placement.conf',
                        '--output-file',
                        '/opt/nova/placement.policy.yaml.sample',
                    ],
                    conf=mock.ANY,
                ),
            ]
        )
